# -*- coding: utf-8 -*-
#
# Copyright (C) 2020, 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, List, Type, Dict, Any, Tuple, Union

from transip import TransIP
from transip.base import ApiObject, ApiService


# Typing alias for the _create_attrs, _update_attrs and _delete_attrs
# attributes in the CreateMixin, UpdateMixin and DeleteMixin
AttrsTuple = Tuple[
    Union[Tuple[()], Tuple[str, ...]],
    Union[Tuple[()], Tuple[str, ...]]
]


class GetMixin:
    """
    Retrieve an single ApiObject.

    Derived class must define ``_resp_get_attr``.

    ``_resp_get_attr``: The response attribute which contains the object
    """
    client: TransIP
    _obj_cls: Optional[Type[ApiObject]]
    path: str

    _resp_get_attr: Optional[str] = None

    def get(self, id: str) -> Optional[Type[ApiObject]]:
        if self._obj_cls or self.path or self._resp_get_attr:
            obj: Type[ApiObject] = self._obj_cls(  # type: ignore
                self,
                self.client.get(f"{self.path}/{id}")[self._resp_get_attr]
            )
            return obj
        return None


class DeleteMixin:
    """Delete a single ApiObject."""

    client: TransIP
    path: str

    def delete(self, id: str) -> None:
        if self.path:
            self.client.delete(f"{self.path}/{id}")


class ObjectDeleteMixin:
    """Delete a single ApiObject."""

    service: ApiService

    def delete(self) -> None:
        if self.get_id():  # type: ignore
            self.service.delete(self.get_id())  # type: ignore


class ObjectUpdateMixin:
    """Update a single ApiObject."""

    service: ApiService
    _updated_attrs: Any

    def _get_updated_data(self) -> Union[Dict, Dict[str, Any]]:
        updated_data = {}
        required, optional = self.service.get_update_attrs()  # type: ignore
        # Add all required attributes, even if they haven't been updated
        for attr in required:
            updated_data[attr] = getattr(self, attr)
        updated_data.update(self._updated_attrs)
        return updated_data

    def update(self) -> None:
        """
        Update the changes made to the object.
        """
        updated_data = self._get_updated_data()
        if not updated_data:
            return

        obj_id = self.get_id()  # type: ignore
        self.service.update(obj_id, updated_data)  # type: ignore


class ListMixin:
    """
    Retrieve a list of ApiObjects.

    Derived class must define ``_resp_list_attr``.

    ``_resp_list_attr``: The response attribute which lists all objects
    """
    client: TransIP
    path: str
    _obj_cls: Optional[Type[ApiObject]]

    _resp_list_attr: Optional[str] = None

    def list(self) -> List[Type[ApiObject]]:
        objs: List[Type[ApiObject]] = []
        if self._obj_cls and self.path and self._resp_list_attr:
            for obj in self.client.get(self.path)[self._resp_list_attr]:
                objs.append(self._obj_cls(self, obj))  # type: ignore
        return objs


class UpdateMixin:
    """
    Update an ApiObject.
    """

    client: TransIP
    path: str

    _req_update_attr: Optional[str] = None
    _update_attrs: Optional[AttrsTuple] = None

    def get_update_attrs(self) -> Tuple[
        Union[Tuple[()], Tuple[str, ...]],
        Union[Tuple[()], Tuple[str, ...]]
    ]:
        """
        Return the required and optional attributes for updating an object.

        Returns:
            tuple: a tuple containing a tuple of required and optional
                attributes.
        """
        if not self._update_attrs:
            return (tuple(), tuple())
        else:
            return self._update_attrs

    def _check_update_attrs(self, attrs) -> None:
        """
        Check required attributes.

        Raises:
            AttributeError: If any of the required attributes is missing.
        """
        required, optional = self.get_update_attrs()
        missing = [attr for attr in required if attr not in attrs]

        if missing:
            attrs_str = "', '".join(missing)
            raise AttributeError((
                f"'{self.__class__.__name__}' object has no "
                f"attribute{'s'[:len(missing)!=1]} '{attrs_str}'"
            ))

    def update(
        self,
        id: Any,
        data: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> None:
        if data is None:
            data = {}

        # Check if all required attributes are supplied
        self._check_update_attrs(data)

        # Some endpoints require the attributes to be packed in dictionary with
        # a specific key while others endpoint may not
        if self._req_update_attr:
            data = {self._req_update_attr: data}

        if self.path:
            self.client.put(f"{self.path}/{id}", json=data)


class ReplaceMixin:
    """
    Replace a list of ApiObject at once by wiping to old objects and replacing
    them with the provided list of ApiObjects.
    """

    client: TransIP
    path: str

    _req_replace_attr: Optional[str] = None
    _replace_attrs: Optional[AttrsTuple] = None

    def get_replace_attrs(self) -> Tuple[
        Union[Tuple[()], Tuple[str, ...]],
        Union[Tuple[()], Tuple[str, ...]]
    ]:
        """
        Return the required and optional attributes for replacing a object.

        Returns:
            tuple: a tuple containing a tuple of required and optional
                attributes.
        """
        if not self._replace_attrs:
            return (tuple(), tuple())
        else:
            return self._replace_attrs

    def replace(self, objs: List[Type[ApiObject]]) -> None:
        """
        Replace all existing objects with the provided once.

        Args:
            objs: List of ApiObjects to replace the existing once with.
        """
        data = []

        required, optional = self.get_replace_attrs()

        for obj in objs:
            obj_data = {}

            # Ensure all required attributes are added
            for attr in required:
                obj_data[attr] = getattr(obj, attr)
            # Ensure all optional attributes are added
            for attr in optional:
                obj_data[attr] = getattr(obj, attr)
            # Overwrite the existing attributes with any updated attributes
            obj_data.update(obj._updated_attrs)  # type: ignore

            data.append(obj_data)

        # Some endpoints require the attributes to be packed in dictionary with
        # a specific key while others endpoint may not
        if self._req_replace_attr:
            data = {self._req_replace_attr: data}  # type: ignore

        if self.path:
            self.client.put(self.path, json=data)


class CreateMixin:
    """
    Create a new ApiObject.
    """
    client: TransIP
    path: str

    _req_create_attr: Optional[str] = None
    _create_attrs: Optional[AttrsTuple] = None

    def _check_create_attrs(self, attrs) -> None:
        """
        Check required attributes.

        Raises:
            AttributeError: If any of the required attributes is missing.
        """
        required, optional = self.get_create_attrs()
        missing = [attr for attr in required if attr not in attrs]

        if missing:
            attrs_str = "', '".join(missing)
            raise AttributeError((
                f"'{self.__class__.__name__}' object has no "
                f"attribute{'s'[:len(missing)!=1]} '{attrs_str}'"
            ))

    def get_create_attrs(self) -> Tuple[
        Union[Tuple[()], Tuple[str, ...]],
        Union[Tuple[()], Tuple[str, ...]]
    ]:
        """
        Return the required and optional attributes for creating a new object.

        Returns:
            tuple: a tuple containing a tuple of required and optional
                attributes.
        """
        if not self._create_attrs:
            return (tuple(), tuple())
        else:
            return self._create_attrs

    def create(self, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = {}

        # Check if all required attributes are supplied
        self._check_create_attrs(data)

        # Some endpoints require the attributes to be packed in dictionary with
        # a specific key while others endpoint do not
        if self._req_create_attr:
            data = {self._req_create_attr: data}

        if self.path:
            self.client.post(self.path, json=data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, List, Dict
import json
import os
import responses  # type: ignore


def load_responses_fixtures(path) -> None:
    """Load a JSON fixture containing all the API response examples."""

    def _load_json_fixtures(path: str) -> List[Dict[str, Any]]:
        """Load JSON fixtures from file."""
        cwd: str = os.path.dirname(os.path.realpath(__file__))
        fixtures: str = os.path.join(os.path.join(cwd, 'fixtures'), path)
        with open(fixtures) as fixture:
            return json.load(fixture)

    def _get_responses_method(method: str) -> str:
        """Returns the responses method based upon the supplied method.

        Args:
            method (str): The response method.

        Raises:
            ValueError: if the specified method is invalid.
        """
        method = method.upper()
        if method == "GET":
            return responses.GET
        elif method == "POST":
            return responses.POST
        elif method == "DELETE":
            return responses.DELETE
        elif method == "PUT":
            return responses.PUT
        elif method == "PATCH":
            return responses.PATCH
        raise ValueError(f"Unable to find method '{method}' in responses")
    
    fixtures: List[Dict[str, Any]] = _load_json_fixtures(path)
    for fixture in fixtures:
        # Add the matchers for the request parameters or JSON body
        matchers: List[Any] = []
        if fixture.get('match_json_params'):
            matchers.append(
                responses.json_params_matcher(fixture['match_json_params'])
            )
        if fixture.get('match_urlencoded_params'):
            matchers.append(
                responses.urlencoded_params_matcher(
                    fixture['match_urlencoded_params']
                )
            )

        # Register the mocked response
        responses.add(
            _get_responses_method(fixture["method"]),
            url=fixture["url"],
            json=fixture.get("json"),
            status=fixture["status"],
            content_type=fixture.get("content_type", "application/json"),
            match=matchers
        )



# Configuration file for the Sphinx documentation builder.
import os
import sys
from datetime import datetime

sys.path.append("../")
sys.path.append(os.path.dirname(__file__))
# sys.path.insert(0, os.path.abspath('..'))

import transip


# Project information
project = transip.__title__
copyright = '{}, {}'.format(
    datetime.now().year,
    transip.__author__
)
author = transip.__author__
version = transip.__version__
release = version

# General configuration
extensions = ['m2r2']
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
source_suffix = ['.rst', '.md']

# Options for HTML output
html_theme = 'alabaster'
html_theme_options = {
    "description": "Python wrapper for the TransIP API. It allows you to create, read and update resources on TransIP with ease using Python.",
    "show_powered_by": False,
    "github_user": "roaldnefs",
    "github_repo": "python-transip",
    "github_banner": True,
    "show_related": False,
    "note_bg": "#FFF59C",
}
html_static_path = ['_static']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020, 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

import responses  # type: ignore
import unittest

from typing import List, Dict, Any, Union

from transip import TransIP
from transip.v6.objects import Domain, WhoisContact, Nameserver, DnsEntry
from tests.utils import load_responses_fixtures


class DomainsTest(unittest.TestCase):
    """Test the DomainService."""

    client: TransIP

    @classmethod
    def setUpClass(cls) -> None:
        """Set up a minimal TransIP client for using the domain services."""
        cls.client = TransIP(access_token='ACCESS_TOKEN')

    def setUp(self) -> None:
        """Setup mocked responses for the '/domains' endpoint."""
        load_responses_fixtures("domains.json")

    @responses.activate
    def test_get(self) -> None:
        domain: Domain = self.client.domains.get("example.com")  # type: ignore

        assert domain.get_id() == "example.com"  # type: ignore

    @responses.activate
    def test_contacts_list(self) -> None:
        domain: Domain = self.client.domains.get("example.com")  # type: ignore
        contacts: List[Domain] = domain.contacts.list()  # type: ignore
        contact: Domain = contacts[0]

        assert len(contacts) == 1
        assert contact.companyName == "Example B.V."  # type: ignore

    @responses.activate
    def test_nameservers_list(self) -> None:
        domain: Domain = self.client.domains.get("example.com")  # type: ignore
        nameservers: List[Nameserver] = domain.nameservers.list()  # type: ignore
        nameserver: Nameserver = nameservers[0]

        assert len(nameservers) == 1
        assert nameserver.get_id() == "ns0.transip.nl"  # type: ignore

    @responses.activate
    def test_dns_list(self) -> None:
        """
        Check if the DNS records for a single domain can be listed.
        """
        domain: Domain = self.client.domains.get("example.com")  # type: ignore
        entries: List[DnsEntry] = domain.dns.list()  # type: ignore

        self.assertEqual(len(entries), 1)
        entry: DnsEntry = entries[0]
        self.assertEqual(entry.content, "127.0.0.1")  # type: ignore

    @responses.activate
    def test_dns_replace(self) -> None:
        """
        Check if the existing DNS records for a single domain can be replaced
        at once.
        """
        domain: Domain = self.client.domains.get("example.com")  # type: ignore
        records: List[DnsEntry] = domain.dns.list()  # type: ignore

        # Ensure we have a single DNS record
        self.assertEqual(len(records), 1)
        # Update the content of the first DNS record in the list of existing
        # records.
        records[0].content = '127.0.0.2'

        # Replace all existing records.
        try:
            domain.dns.replace(records)  # type: ignore
        except Exception as exc:
            assert False, f"'transip.v6.objects.Domain.dns.replace' raised an exception {exc}"


    @responses.activate
    def test_dns_create(self) -> None:
        dns_entry_data: Dict[str, Union[str, int]] = {
            "name": "www",
            "expire": 86400,
            "type": "A",
            "content": "127.0.0.1"
        }
        domain: Domain = self.client.domains.get("example.com")  # type: ignore
        domain.dns.create(dns_entry_data)  # type: ignore

        assert len(responses.calls) == 2

    @responses.activate
    def test_dns_update(self) -> None:
        """
        Check if a single DNS entry can be updated.
        """
        dns_entry_data: Dict[str, Union[str, int]] = {
            "name": "www",
            "expire": 86400,
            "type": "A",
            "content": "127.0.0.2"  # original content is 127.0.0.1
        }
        domain: Domain = self.client.domains.get("example.com")  # type: ignore

        try:
            domain.dns.update(dns_entry_data)
        except Exception as exc:
            assert False, f"'transip.v6.objects.Domain.dns.update' raised an exception {exc}"

    @responses.activate
    def test_dns_update_object(self) -> None:
        """
        Check if a single DNS entry can be updated from the ApiObject itself.
        """
        domain: Domain = self.client.domains.get("example.com")  # type: ignore
        entries: List[DnsEntry] = domain.dns.list()  # type: ignore

        self.assertEqual(len(entries), 1)
        entry: DnsEntry = entries[0]
        self.assertEqual(entry.content, "127.0.0.1")  # type: ignore

        entry.content = '127.0.0.2'
        try:
            entry.update()
        except Exception as exc:
            assert False, f"'transip.v6.objects.DnsEntry.update' raised an exception {exc}"

        self.assertEqual(entry.content, "127.0.0.2")  # type: ignore

    @responses.activate
    def test_dns_delete(self) -> None:
        """Check if a single DNS entry can be deleted."""
        dns_entry_data: Dict[str, Union[str, int]] = {
            "name": "www",
            "expire": 86400,
            "type": "A",
            "content": "127.0.0.1"
        }
        domain: Domain = self.client.domains.get("example.com")  # type: ignore

        try:
            domain.dns.delete(dns_entry_data)  # type: ignore
        except Exception as exc:
            assert False, (
                "'transip.v6.objects.Domain.dns.delete' raised an exception "
                f"{exc}"
            )
        # Ensure only two calls are being made: the retrieval of the domain and
        # the deletion of the DNS entry
        self.assertEqual(len(responses.calls), 2)

    @responses.activate
    def test_dns_delete_object(self) -> None:
        """Check if a DNS entry can be deleted from the object itself."""
        domain: Domain = self.client.domains.get("example.com")  # type: ignore
        entries = domain.dns.list()

        self.assertEqual(len(entries), 1)

        entry: DnsEntry = entries[0]  # type: ignore
        try:
            # Delete the first DNS entry
            entry.delete()
        except Exception as exc:
            assert False, (
                "'transip.v6.objects.DnsEntry.delete' raised an exception "
                f"{exc}"
            )
        # Ensure only three calls are being made; the retrieval of the domain,
        # the listing of the DNS entries and the deletion of a single DNS
        # entry.
        self.assertEqual(len(responses.calls), 3)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020, 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.
"""Wrapper for the TransIP API."""

from typing import Dict, Optional, Any, Type, Union, TYPE_CHECKING
from types import ModuleType

import importlib
import requests
import os

from transip.exceptions import TransIPHTTPError, TransIPParsingError
from transip.utils import generate_message_signature, generate_nonce


__title__ = "python-transip"
__version__ = "0.4.0"
__author__ = "Roald Nefs"
__email__ = "info@roaldnefs.com"
__copyright__ = "Copyright 2020-2021, Roald Nefs"
__license__ = "LGPL3"


if TYPE_CHECKING:
    # Imports only needed for type checking. These will not be imported at
    # runtime.
    from transip.base import ApiService


class TransIP:
    """Represents a TransIP server connection.

    Args:
        login (str): The TransIP username
        api_version (str): TransIP API version to use
        access_token (str): The TransIP API access token
        private_key (str): The content of the private key for accessing the
            TransIP API
        private_key_file (str): Path to the private key for accessing the
            TransIP API
    """

    def __init__(
        self,
        login: str = None,
        api_version: str = "6",
        access_token: Optional[str] = None,
        private_key: Optional[str] = None,
        private_key_file: Optional[str] = None,
    ) -> None:
        self._api_version: str = api_version
        self._url: str = f"https://api.transip.nl/v{api_version}"

        # Headers to use when making a request to TransIP
        self.headers: Dict[str, str] = {
            "User-Agent": f"{__title__}/{__version__}"
        }

        # Initialize a session object for making requests
        self.session: requests.Session = requests.Session()

        # Set authentication information
        self._login: Optional[str] = login
        self._access_token: Optional[str] = access_token
        self._private_key: Optional[str] = private_key
        self._private_key_file: Optional[str] = private_key_file
        self._set_auth_info()

        # Dynamically import the services for the specified API version
        objects: ModuleType = (
            importlib.import_module(f"transip.v{api_version}.objects")
        )

        self.api_test: Type['ApiService'] = (
            objects.ApiTestService(self)  # type: ignore
        )
        self.availability_zones: Type['ApiService'] = (
            objects.AvailabilityZoneService(self)  # type: ignore
        )
        self.products: Type['ApiService'] = (
            objects.ProductService(self)  # type: ignore
        )
        self.domains: Type['ApiService'] = (
            objects.DomainService(self)  # type: ignore
        )
        self.invoices: Type['ApiService'] = (
            objects.InvoiceService(self)  # type: ignore
        )
        self.ssh_keys: Type['ApiService'] = (
            objects.SshKeyService(self)  # type: ignore
        )
        self.vpss: Type['ApiService'] = (
            objects.VpsService(self)  # type: ignore
        )

    @property
    def url(self) -> str:
        """Return the API URL."""
        return self._url

    def _get_headers(
        self,
        content_type: Optional[str] = None
    ) -> Dict[str, str]:
        headers = self.headers.copy()
        if content_type:
            headers["Content-Type"] = content_type
        return headers

    def _build_url(self, path: str) -> str:
        return f"{self._url}{path}"

    def _request_access_token(self) -> str:
        """
        Request an access token using the supplied private key.

        Returns:
            str: The access token to use for authorization.

        Raises:
            TransIPParsingError: If the requested access token couldn't be
                extracted from the API response.
        """
        url: str = self._build_url('/auth')
        payload: Dict[str, Any] = {
            "login": self._login,
            # The TransIP API requires that the length of the nonce is between
            # 6 and 32 characters
            "nonce": generate_nonce(32),
            # TODO(roaldnefs): Allow the creation of read-only access tokens
            "read_only": False,
            # TODO(roaldnefs): Allow the expiration time of the access token
            # to be overwritten
            # "expiration_time": "30 minutes",
            # TODO(roaldnefs): Allow a custom label to be specified when
            # generating a new access token
            # "label": "python-transip",
            # TODO(roaldnefs): Allow the access token to only be use from
            # whitelisted IP-addresses
            "global_key": False
        }

        headers: Dict[str, str] = self.headers.copy()
        request: requests.Request = requests.Request(
            "POST", url, headers=headers, json=payload
        )
        prepped: requests.PreparedRequest = self.session.prepare_request(
            request
        )

        # Get the prepped body for signature generation
        body: Union[bytes, str] = prepped.body or ''
        if isinstance(body, bytes):
            body = body.decode('ascii')

        # Generate a signature if the request body
        signature: str = generate_message_signature(
            body, self._private_key  # type: ignore
        )

        # Add 'Signature' header to the prepared request
        prepped.headers["Signature"] = signature

        response: requests.Response = self.session.send(prepped)
        data = self._validate_response(response)

        # Attempt to extract the access token from the result
        try:
            return data['token']
        except (AttributeError, KeyError) as exc:
            raise TransIPParsingError(
                "Failed to extract access token from the API response"
            ) from exc

    def _read_private_key(self) -> str:
        """Read the private key from file.

        Returns:
            str: The private key content

        Raises:
            RuntimeError: If the private key file doesn't exist
        """
        if os.path.exists(self._private_key_file):  # type: ignore
            try:
                with open(self._private_key_file) as keyfile:  # type: ignore
                    return keyfile.read()
            except IOError as exc:
                raise RuntimeError("The private key couldn't be read") from exc
        else:
            raise RuntimeError("The private key doesn't exist")

    def _set_auth_info(self) -> None:
        """
        Set authentication information based upon the defined attributes.

        Raises:
            ValueError: If the required attributes are not defined.
        """
        if (not self._access_token and not self._private_key and not
                self._private_key_file):
            raise ValueError(
                "At least one of access_token, private_key and "
                "private_key_file should be defined"
            )
        if self._access_token and self._private_key:
            raise ValueError(
                "Only one of access_token and private_key should be defined"
            )
        if self._access_token and self._private_key_file:
            raise ValueError(
                "Only one of access_token and private_key_file should be "
                "defined"
            )
        if self._private_key and self._private_key_file:
            raise ValueError(
                "Only one of private_key and private_key_file should be "
                "defined"
            )
        if self._private_key and not self._login:
            raise ValueError(
                "Both private_key and login should be defined"
            )
        if self._private_key_file and not self._login:
            raise ValueError(
                "Both private_key_file and login should be defined"
            )

        # Read the private key from file
        if self._private_key_file:
            self._private_key = self._read_private_key()

        # Use the private key to request a new access token
        if self._private_key:
            self._access_token = self._request_access_token()

        # Set the 'Authorization' header
        self.headers["Authorization"] = f"Bearer {self._access_token}"

    def request(
        self,
        method: str,
        path: str,
        data: Optional[Any] = None,
        json: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Make an HTTP request to the TransIP API.

        Args:
            method (str): HTTP method to use
            path (str): The path to append to the API URL
            data (dict): The body to attach to the request
            json (dict): The json body to attach to the request
            params (dict): URL parameters to append to the URL

        Returns:
            Returns the json-encoded content of a response, if any.

        Raises:
            TransIPHTTPError: When the return code of the request is not 2xx
            TransIPParsingError: When the content couldn't be parsed as JSON
        """
        url: str = self._build_url(path)

        # Set the content type for the request if json is provided and data is
        # not specified
        content_type: Optional[str] = None
        if not data and json:
            content_type = "application/json"

        headers: Dict[str, str] = self._get_headers(content_type)
        request: requests.Request = requests.Request(
            method, url, headers=headers, data=data, json=json, params=params,
        )

        prepped: requests.PreparedRequest = self.session.prepare_request(
            request
        )
        response: requests.Response = self.session.send(prepped)
        return self._validate_response(response)

    def _validate_response(self, response: requests.Response) -> Any:
        """
        Validate the API response.

        Raises:
            TransIPHTTPError: When the return code of the request is not 2xx
            TransIPParsingError: When the content couldn't be parsed as JSON
        """
        if 200 <= response.status_code < 300:
            if response.text:
                try:
                    return response.json()
                except Exception:
                    raise TransIPParsingError(
                        message="Failed to parse the API response as JSON"
                    )
            return None

        error_message = str(response.content)
        try:
            error_json = response.json()
            if "error" in error_json:
                error_message = error_json["error"]
        except (KeyError, ValueError, TypeError):
            pass

        raise TransIPHTTPError(
            message=error_message,
            response_code=response.status_code
        )

    def get(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Make a GET request to the TransIP API.

        Args:
            path (str): The path to append to the API URL
            params (dict): URL parameters to append to the URL

        Returns:
            Returns the json-encoded content of a response, if any.

        Raises:
            TransIPHTTPError: When the return code of the request is not 2xx
        """
        return self.request(
            "GET", path, params=params
        )

    def post(
        self,
        path: str,
        data: Optional[Any] = None,
        json: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Make a POST request to the TransIP API.

        Args:
            path (str): The path to append to the API URL
            data (dict): The body to attach to the request
            json (dict): The json body to attach to the request
            params (dict): URL parameters to append to the URL

        Returns:
            Returns the json-encoded content of a response, if any.

        Raises:
            TransIPHTTPError: When the return code of the request is not 2xx
        """
        return self.request(
            "POST", path, data=data, json=json, params=params
        )

    def put(
        self,
        path: str,
        data: Optional[Any] = None,
        json: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Make a PUT request to the TransIP API.

        Args:
            path (str): The path to append to the API URL
            data (dict): The body to attach to the request
            json (dict): The json body to attach to the request
            params (dict): URL parameters to append to the URL

        Returns:
            Returns the json-encoded content of a response, if any.

        Raises:
            TransIPHTTPError: When the return code of the request is not 2xx
        """
        return self.request(
            "PUT", path, data=data, json=json, params=params
        )

    def patch(
        self,
        path: str,
        data: Optional[Any] = None,
        json: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Make a PATCH request to the TransIP API.

        Args:
            path (str): The path to append to the API URL
            data (dict): The body to attach to the request
            json (dict): The json body to attach to the request
            params (dict): URL parameters to append to the URL

        Returns:
            Returns the json-encoded content of a response, if any.

        Raises:
            TransIPHTTPError: When the return code of the request is not 2xx
        """
        return self.request(
            "PATCH", path, data=data, json=json, params=params
        )

    def delete(
        self,
        path: str,
        data: Optional[Any] = None,
        json: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Make a GET request to the TransIP API.

        Args:
            path (str): The path to append to the API URL
            data (dict): The body to attach to the request
            json (dict): The json body to attach to the request
            params (dict): URL parameters to append to the URL

        Returns:
            Returns the json-encoded content of a response, if any.

        Raises:
            TransIPHTTPError: When the return code of the request is not 2xx
        """
        return self.request(
            "DELETE", path, data=data, json=json, params=params
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

import unittest
import string

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from transip.utils import (
    load_rsa_private_key, generate_message_signature, generate_nonce
)


class UtilsTest(unittest.TestCase):
    """Test the transip.utils functions."""

    privkey: str

    @classmethod
    def setUpClass(cls) -> None:
        cls.privkey: str = (  # type: ignore
            "-----BEGIN RSA PRIVATE KEY-----\n"
            "MIIEpAIBAAKCAQEAsUSEHsMuB380OUZQWDyyND4q8lEuJAgNnMkO8s5NGwzP8XSi\n"
            "2DdFglLGLe9kjpADs3XqZFsk8ZFFn7x0idFydGyh9tbJ2WkR9E+kNUJV5iQDzPOB\n"
            "wvyygEREqnl/o1h3c1q8tD2HZKBcjChn9JbMzdWwAaIs3ppcGWrEI0jZFFfSAyIZ\n"
            "GkC3k3umOykWIKflQcT/soAfdqW+2P9/KD/wb3AZCer2i6B2hiITiDbHh5q84Hgk\n"
            "D/Zg1M4yrYDyxDeGkAJHkGKNaE0tgUPoz3XTGP7uFYIx00qJyhmnzQcyV/Xcw3ZQ\n"
            "7DFUj1HQ5wG/kEF9a4F1+AAiO5C5QbGTFYSwBwIDAQABAoIBABbtIZlI7P8TOJHf\n"
            "wixnTTTshWlpjmoikIAikMheXiKNeadkylrkaxz7z53JRFwbzB69tV7dWt3TSAns\n"
            "ubXJXOAp3JisFtcDe8r5MeeheLKXHda396RcQknMioTxycw6eNh2d8ln28br5oxJ\n"
            "/YfoqPxGEsljTCJOHHM9F7johwrWSQ6f+gmiOkABvIHKgTBLa++v0D+vNrUjM6rx\n"
            "IE+dBrx8yIgkF4qSg4Dqnr7D0KqCZUGLZ/3K8ShQUtiQYzyHIWKUId3NUecIQcrT\n"
            "2Ri2TITKuER0fa7Mr+3LMSh/3+HtP2AoM34ouxr9H98LFz/UXxuFIRFTx7UVRt4N\n"
            "3zqhsEECgYEA+TnXanBJmFz3sNYtlQixtKrh496GB0NheuK4xeNEj9/3gJ6J/rtL\n"
            "ZHI7VH8r6aqoqw7sO/WJdxkwZTBOz2fe1QJ5BN0HBI5S6jIBQv9Nfqar0TDvNLB+\n"
            "pH6eYJZ/IEFIMObv9YmsPohXpGeXynecrpl8SazEIWLb8IzgLY0HpokCgYEAthX5\n"
            "1th4Re0P9rzXp21bbEwcvOKcg5dcpSaTtA1eQEILl6qqT3FP7w8/Ed7NRRY9Gcs+\n"
            "inAc96YRNAgIGgfT3R1BmxOMWfdFBT1zlCheS6egKKLzVPzKPiMoMP4zu4hy6uH5\n"
            "YVqpDLu0YQu1J2L0VYdZ9xAC0//Rx8KRcs6m/g8CgYEA41VDja+HMhf7R67WPU+E\n"
            "6YvGKRjdoNpxnKoaaUd5TtO46/WxYk5t4t3gCJ9H6wjkecRO8BJ0pdKwNlzuRno0\n"
            "5JAw26LRt/Iq571dMUO36IMXzuWYDLPBkUJ+LRSaOU3TD+hXkd1W5GNxrmFgMCsT\n"
            "HKCcooeZD+shPDcEdghipiECgYARBTDbYlSrxKMPX0uRPOmkz+CHz27t5gIk9dws\n"
            "omtC+ml2/d75mg/surIci4UIhjGj7Zmk+yHaDE3jXTTUqhKlwoxVYJhn+HMdMEdT\n"
            "fAqEa+DOq5yvPwnwkPy6x6gySWjkh8b10LGonQsZXyzJx7grHoHMVFTPWERVtdw+\n"
            "rQ5zBQKBgQC0Iwx4eeQYx60OCZpioNEQ3QPaFgqoWYEexmcMlpgQ9ycdnHx3SkE8\n"
            "SlMokcPIEJDhdF3632kIAHOOJeA4Tmshf+ol/O2U2PDgbJZL6W6FJlT28sZVUU8j\n"
            "IjFmiAiW6IIEqkJxuR1diAjppEiMmSkjPavo7oQs0TZMMUkli1N9dw==\n"
            "-----END RSA PRIVATE KEY-----"
        )

    def test_load_rsa_private_key(self) -> None:
        """
        Test the content of a RSA private key can be converted to a
        RSAPrivateKey instance.
        """
        self.assertTrue(isinstance(load_rsa_private_key(self.privkey), RSAPrivateKey))

    def test_generate_message_signature(self) -> None:
        """
        Test message signature generation using the defined RSA private key.

        Example message encoded using:
        https://8gwifi.org/rsasignverifyfunctions.jsp
        """
        message: str = "A message for signing"
        encoded: str = ("NFi2v07lhYmyTarOtIfpw50W25ukKWjtqsVzti/Y2RiGKPEzJQtFZ"
                        "QaYJCFfIn8HfYjdbzOTK5DIFxwL8NCJK3Mb+wxZOkO4NDJC7mVgdO"
                        "I6VuET4F3Er4ZjO4pkMLSaV6B0Mcm/yj8Wom1lfeRZxItDXPAbkMj"
                        "47Ywsx7enAEXfrZrYwHy+rWLPN6WWCrCDWAJGu7lz5+YIy7rpLyRx"
                        "Ff57QkMMJal0VCWyQUx+JBMdoW7rGVN1u+AxRY0yFj+QxWRB1z0JC"
                        "E0Xmur+gQ+4+rgIEDE6VU2VY0A8+SY7hyRb2JN8yoLAeI+21ODwo5"
                        "h/x1zw3Bstyzuvzo0QmHp7Mw==")

        self.assertTrue(
            generate_message_signature(message, self.privkey) == encoded
        )

    def test_generate_nonce_length(self) -> None:
        """
        Test the length of the generated nonce and whether or not an
        exception is thrown for invalid lengths.
        """
        # Test valid lengths
        for length in [1, 2, 32]:
            self.assertTrue(len(generate_nonce(length)) == length)

        # Test invalid lengths
        for length in [0, -1]:
            self.assertRaises(ValueError, generate_nonce, length)

    def test_generate_nonce_alphabet(self) -> None:
        """
        Test if the generated nonce only contains characters from the alphabet.
        """
        alphabet: str = 'a'
        self.assertTrue(generate_nonce(3, alphabet) == 'aaa')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union

import base64
import secrets
import string

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.hashes import SHA512
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15


def load_rsa_private_key(key: Union[bytes, str]) -> RSAPrivateKey:
    """
    Convert the private key string to RSAPrivateKey object.

    Returns:
        RSAPrivateKey: The private RSA key.
    """
    # Convert the key string to bytes
    if isinstance(key, str):
        key = key.encode()

    return serialization.load_pem_private_key(
        key, password=None, backend=default_backend()
    )


def generate_message_signature(
    message: Union[str, bytes],
    private_key: Union[RSAPrivateKey, str]
) -> str:
    """Return the BASE64 encoded SHA514 signature of a message.

    Args:
        message (str): The message to sign.
        private_key (str): The private key content used to sign the message.

    Returns:
        str: The BASE64 encoded SHA514 signature of a message.
    """
    # Convert the message string to bytes
    if isinstance(message, str):
        message = message.encode()

    # Convert the private key content to a RSAPrivateKey object
    if isinstance(private_key, str):
        private_key = load_rsa_private_key(private_key)

    # Sign the message using the RSAPrivateKey object
    signature: bytes = private_key.sign(message, PKCS1v15(), SHA512())

    # Return the BASE64 encoded SHA512 signature
    b64_bytes: bytes = base64.b64encode(signature)
    return b64_bytes.decode('ascii')


def generate_nonce(length: int, alphabet: str = None) -> str:
    """
    Generate a nonce.

    Args:
        length (int): The number of characters to return.
        alphabet (str): The alphabet to choose characters from, defaults to
            ascii leters and digits.

    Returns:
        str: The nonce of specified characters.
    """
    if not length >= 1:
        raise ValueError(
            "The specified nonce length must greater or equal to 1"
        )

    alphabet = alphabet or (string.ascii_letters + string.digits)
    return ''.join(secrets.choice(alphabet) for i in range(length))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020, 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Type, Any, Union

from transip import TransIP


class ApiObject:
    """Represents a TransIP API object."""

    _id_attr: Optional[str] = "id"

    def __init__(self, service, attrs) -> None:
        self.__dict__.update(
            {
                "service": service,
                "_attrs": attrs,
                "_updated_attrs": {}
            }
        )

    def __getattr__(self, name: str) -> Any:
        try:
            return self.__dict__["_updated_attrs"][name]
        except KeyError:
            try:
                return self.__dict__["_attrs"][name]
            except KeyError:
                raise AttributeError(name)

    def __setattr__(self, name: str, value: Any) -> None:
        self.__dict__["_updated_attrs"][name] = value

    def __str__(self) -> str:
        return f"{type(self)} => {self._attrs}"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        if self._id_attr:
            return f"<{name} {self._id_attr}:{self.get_id()}>"
        else:
            return f"<{name}>"

    def __dir__(self):
        return super().__dir__() + list(self.attrs)

    def get_id(self) -> Union[Optional[int], Optional[str]]:
        """Returns the ID of the object."""
        if self._id_attr and hasattr(self, self._id_attr):
            return getattr(self, self._id_attr)
        return None

    @property
    def attrs(self):
        """
        Returns a dictionary containing all the attributes.
        """
        attrs = self.__dict__["_updated_attrs"].copy()
        attrs.update(self.__dict__["_attrs"])
        return attrs


class ApiService:
    """Represents a TransIP API service."""

    _path: Optional[str] = None
    _obj_cls: Optional[Type[ApiObject]] = None

    def __init__(
        self,
        client: TransIP,
        parent: Optional[Type[ApiObject]] = None
    ) -> None:
        self.client: TransIP = client
        self._parent: Optional[Type[ApiObject]] = parent

    @property
    def path(self) -> Optional[str]:
        if self._path and self._parent:
            return self._path.format(
                parent_id=self._parent.get_id()  # type: ignore
            )
        return self._path



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

import responses  # type: ignore
import unittest

from transip import TransIP

from tests.utils import load_responses_fixtures


class ApiTestTest(unittest.TestCase):
    """Test the ApiTestService."""

    client: TransIP

    @classmethod
    def setUpClass(cls) -> None:
        """Set up a minimal TransIP client for using the API test services."""
        cls.client = TransIP(access_token='ACCESS_TOKEN')

    def setUp(self) -> None:
        """Setup mocked responses for the '/api-test' endpoint."""
        load_responses_fixtures("general.json")

    @responses.activate
    def test_test(self) -> None:
        """Test if the API test returns True."""
        self.assertTrue(self.client.api_test.test())  # type: ignore



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020, 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

import unittest
import responses  # type: ignore

from transip import TransIP
from tests.utils import load_responses_fixtures


class TransIPTest(unittest.TestCase):
    """Test the TransIP client class."""

    client: TransIP
    privkey: str

    @classmethod
    def setUpClass(cls) -> None:
        """Set up a minimal TransIP client and RSA private key."""
        cls.client = TransIP(access_token='ACCESS_TOKEN')
        cls.privkey: str = (  # type: ignore
            "-----BEGIN RSA PRIVATE KEY-----\n"
            "MIIEpAIBAAKCAQEAsUSEHsMuB380OUZQWDyyND4q8lEuJAgNnMkO8s5NGwzP8XSi\n"
            "2DdFglLGLe9kjpADs3XqZFsk8ZFFn7x0idFydGyh9tbJ2WkR9E+kNUJV5iQDzPOB\n"
            "wvyygEREqnl/o1h3c1q8tD2HZKBcjChn9JbMzdWwAaIs3ppcGWrEI0jZFFfSAyIZ\n"
            "GkC3k3umOykWIKflQcT/soAfdqW+2P9/KD/wb3AZCer2i6B2hiITiDbHh5q84Hgk\n"
            "D/Zg1M4yrYDyxDeGkAJHkGKNaE0tgUPoz3XTGP7uFYIx00qJyhmnzQcyV/Xcw3ZQ\n"
            "7DFUj1HQ5wG/kEF9a4F1+AAiO5C5QbGTFYSwBwIDAQABAoIBABbtIZlI7P8TOJHf\n"
            "wixnTTTshWlpjmoikIAikMheXiKNeadkylrkaxz7z53JRFwbzB69tV7dWt3TSAns\n"
            "ubXJXOAp3JisFtcDe8r5MeeheLKXHda396RcQknMioTxycw6eNh2d8ln28br5oxJ\n"
            "/YfoqPxGEsljTCJOHHM9F7johwrWSQ6f+gmiOkABvIHKgTBLa++v0D+vNrUjM6rx\n"
            "IE+dBrx8yIgkF4qSg4Dqnr7D0KqCZUGLZ/3K8ShQUtiQYzyHIWKUId3NUecIQcrT\n"
            "2Ri2TITKuER0fa7Mr+3LMSh/3+HtP2AoM34ouxr9H98LFz/UXxuFIRFTx7UVRt4N\n"
            "3zqhsEECgYEA+TnXanBJmFz3sNYtlQixtKrh496GB0NheuK4xeNEj9/3gJ6J/rtL\n"
            "ZHI7VH8r6aqoqw7sO/WJdxkwZTBOz2fe1QJ5BN0HBI5S6jIBQv9Nfqar0TDvNLB+\n"
            "pH6eYJZ/IEFIMObv9YmsPohXpGeXynecrpl8SazEIWLb8IzgLY0HpokCgYEAthX5\n"
            "1th4Re0P9rzXp21bbEwcvOKcg5dcpSaTtA1eQEILl6qqT3FP7w8/Ed7NRRY9Gcs+\n"
            "inAc96YRNAgIGgfT3R1BmxOMWfdFBT1zlCheS6egKKLzVPzKPiMoMP4zu4hy6uH5\n"
            "YVqpDLu0YQu1J2L0VYdZ9xAC0//Rx8KRcs6m/g8CgYEA41VDja+HMhf7R67WPU+E\n"
            "6YvGKRjdoNpxnKoaaUd5TtO46/WxYk5t4t3gCJ9H6wjkecRO8BJ0pdKwNlzuRno0\n"
            "5JAw26LRt/Iq571dMUO36IMXzuWYDLPBkUJ+LRSaOU3TD+hXkd1W5GNxrmFgMCsT\n"
            "HKCcooeZD+shPDcEdghipiECgYARBTDbYlSrxKMPX0uRPOmkz+CHz27t5gIk9dws\n"
            "omtC+ml2/d75mg/surIci4UIhjGj7Zmk+yHaDE3jXTTUqhKlwoxVYJhn+HMdMEdT\n"
            "fAqEa+DOq5yvPwnwkPy6x6gySWjkh8b10LGonQsZXyzJx7grHoHMVFTPWERVtdw+\n"
            "rQ5zBQKBgQC0Iwx4eeQYx60OCZpioNEQ3QPaFgqoWYEexmcMlpgQ9ycdnHx3SkE8\n"
            "SlMokcPIEJDhdF3632kIAHOOJeA4Tmshf+ol/O2U2PDgbJZL6W6FJlT28sZVUU8j\n"
            "IjFmiAiW6IIEqkJxuR1diAjppEiMmSkjPavo7oQs0TZMMUkli1N9dw==\n"
            "-----END RSA PRIVATE KEY-----"
        )

    def setUp(self) -> None:
        """Setup mocked responses for the '/auth' endpoint."""
        load_responses_fixtures("auth.json")

    def test_base_url(self) -> None:
        """Test the base URL of the API."""
        url: str = self.client.url

        assert url == "https://api.transip.nl/v6"

    def test_authorization_header(self) -> None:
        """Test if the Authorization header is set correctly."""
        auth_header: str = self.client.headers["Authorization"]

        assert auth_header == "Bearer ACCESS_TOKEN"

    @responses.activate
    def test_private_key_authorization_header(self) -> None:
        """
        Test if TransIP instances initialized with a private key, results in
        the 'Authorization' header containing the access token returned by the
        mocked response.
        """
        client: TransIP = TransIP(login="testuser", private_key=self.privkey)
        auth_header: str = self.client.headers["Authorization"]

        # Assert a single request is made to the mocked TransIP API
        self.assertEqual(len(responses.calls), 1)
        # Assert the 'Authorization' header contains the access token returned
        # by the mocked response
        self.assertEqual(auth_header, "Bearer ACCESS_TOKEN")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

# The demo token provided by TransIP to authenticate as the TransIP demo user
# in test mode.
DEMO_TOKEN = 'eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiIsImp0aSI6ImN3MiFSbDU2eDNoUnkjelM4YmdOIn0.eyJpc3MiOiJhcGkudHJhbnNpcC5ubCIsImF1ZCI6ImFwaS50cmFuc2lwLm5sIiwianRpIjoiY3cyIVJsNTZ4M2hSeSN6UzhiZ04iLCJpYXQiOjE1ODIyMDE1NTAsIm5iZiI6MTU4MjIwMTU1MCwiZXhwIjoyMTE4NzQ1NTUwLCJjaWQiOiI2MDQ0OSIsInJvIjpmYWxzZSwiZ2siOmZhbHNlLCJrdiI6dHJ1ZX0.fYBWV4O5WPXxGuWG-vcrFWqmRHBm9yp0PHiYh_oAWxWxCaZX2Rf6WJfc13AxEeZ67-lY0TA2kSaOCp0PggBb_MGj73t4cH8gdwDJzANVxkiPL1Saqiw2NgZ3IHASJnisUWNnZp8HnrhLLe5ficvb1D9WOUOItmFC2ZgfGObNhlL2y-AMNLT4X7oNgrNTGm-mespo0jD_qH9dK5_evSzS3K8o03gu6p19jxfsnIh8TIVRvNdluYC2wo4qDl5EW5BEZ8OSuJ121ncOT1oRpzXB0cVZ9e5_UVAEr9X3f26_Eomg52-PjrgcRJ_jPIUYbrlo06KjjX2h0fzMr21ZE023Gw'  # noqa E501



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Tuple, Any, Dict, Optional
import responses  # type: ignore
import unittest
import tempfile
import os

from transip import TransIP
from transip.v6.objects import Product, ProductElement
from tests.utils import load_responses_fixtures


class ProductsTest(unittest.TestCase):
    """Test the ProductsService."""

    client: TransIP

    @classmethod
    def setUpClass(cls) -> None:
        """Set up a minimal TransIP client for using the invoice services."""
        cls.client = TransIP(access_token='ACCESS_TOKEN')

    def setUp(self) -> None:
        """Setup mocked responses for the '/products' endpoint."""
        load_responses_fixtures("general.json")

    @responses.activate
    def test_list(self) -> None:
        """Check if the products can be listed."""
        products: List[Product] = self.client.products.list()  # type: ignore
    
        self.assertEqual(len(products), 5)

    @responses.activate
    def test_elements_list(self) -> None:
        """Check if the elements of a product can be listed."""
        products: List[Product] = self.client.products.list()  # type: ignore
        product: Optional[Product] = None

        # Find the correct product as the API doesn't provide an option to
        # retrieve a single product
        for entry in products:
            if entry.get_id() == 'vps-bladevps-x4':
                product = entry
                break
        
        self.assertIsNotNone(product)

        elements = product.elements.list()  # type: ignore
        self.assertEqual(len(elements), 1)
        self.assertEqual(elements[0].get_id(), 'ipv4Addresses')  # type: ignore



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020, 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Tuple, Any, Dict, Optional
import responses  # type: ignore
import unittest
import tempfile
import os

from transip import TransIP
from transip.v6.objects import Invoice, InvoiceItem
from tests.utils import load_responses_fixtures

class InvoicesTest(unittest.TestCase):
    """Test the InvoiceService."""

    client: TransIP
    tmp_dir: Optional[tempfile.TemporaryDirectory]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up a minimal TransIP client for using the invoice services."""
        cls.client = TransIP(access_token='ACCESS_TOKEN')

    def setUp(self) -> None:
        """Setup mocked responses for the '/invoices' endpoint."""
        load_responses_fixtures("account.json")

    @responses.activate
    def test_list(self) -> None:
        invoices: List[Invoice] = self.client.invoices.list()  # type: ignore
        invoice: Invoice = invoices[0]
    
        self.assertEqual(len(invoices), 1)
        self.assertEqual(invoice.get_id(), "F0000.1911.0000.0004")  # type: ignore

    @responses.activate
    def test_get(self) -> None:
        invoice_id: str = "F0000.1911.0000.0004"
        invoice: Invoice = self.client.invoices.get(invoice_id)  # type: ignore

        self.assertEqual(invoice.get_id(), "F0000.1911.0000.0004")  # type: ignore

    @responses.activate
    def test_items_list(self) -> None:
        invoice: Invoice = self.client.invoices.get("F0000.1911.0000.0004")  # type: ignore
        items: List[InvoiceItem] = invoice.items.list()  # type: ignore

        self.assertEqual(len(items), 1)
        self.assertEqual(items[0].product, "Big Storage Disk 2000 GB")  # type: ignore

        # Expect the get_id() method to return None as the invoice items don't
        # have a specific ID attribute.
        self.assertIsNone(items[0].get_id())  # type: ignore

    @responses.activate
    def test_pdf(self) -> None:
        """
        Check if the invoice can be written to a PDF file.
        """
        invoice_id = "F0000.1911.0000.0004"
        invoice: Invoice = self.client.invoices.get(invoice_id)  # type: ignore

        # Write the invoice to a PDF file in a temporary directory.
        with tempfile.TemporaryDirectory() as tmp_dir:
            expected = os.path.join(tmp_dir, f"{invoice_id}.pdf")
            actual = invoice.pdf(tmp_dir)
            self.assertEqual(actual, expected)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020, 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Tuple, Any, Dict
import responses  # type: ignore
import unittest

from transip import TransIP
from transip.v6.objects import SshKey
from tests.utils import load_responses_fixtures


class SshKeysTest(unittest.TestCase):
    """Test the SshKeyService."""

    client: TransIP

    @classmethod
    def setUpClass(cls) -> None:
        """Set up a minimal TransIP client for using the ssh-key services."""
        cls.client = TransIP(access_token='ACCESS_TOKEN')

    def setUp(self) -> None:
        """Setup mocked responses for the '/ssh-keys' endpoint."""
        load_responses_fixtures("account.json")

    @responses.activate
    def test_list(self) -> None:
        ssh_keys: List[SshKey] = self.client.ssh_keys.list()  # type: ignore
        ssh_key: SshKey = ssh_keys[0]

        assert len(ssh_keys) == 1
        assert ssh_key.get_id() == 123  # type: ignore

    @responses.activate
    def test_get(self) -> None:
        ssh_key_id: int = 123
        ssh_key: SshKey = self.client.ssh_keys.get(ssh_key_id)  # type: ignore

        assert ssh_key.get_id() == 123  # type: ignore

    @responses.activate
    def test_delete(self) -> None:
        ssh_key_id: int = 123
        try:
            self.client.ssh_keys.delete(ssh_key_id)  # type: ignore
        except Exception as exc:
            assert False, f"'transip.TransIP.ssh_keys.delete' raised an exception {exc}"

    @responses.activate
    def test_delete_object(self) -> None:
        ssh_key: SshKey = self.client.ssh_keys.get(123)  # type: ignore

        try:
            ssh_key.delete()  # type: ignore
        except Exception as exc:
            assert False, f"'transip.v6.objects.SshKey.delete' raised an exception {exc}"

    @responses.activate
    def test_update_object(self) -> None:
        ssh_key: SshKey = self.client.ssh_keys.get(123)  # type: ignore
        ssh_key.description = "Jim key"

        try:
            ssh_key.update()
        except Exception as exc:
            assert False, f"'transip.v6.objects.SshKey.update' raised an exception {exc}"

    @responses.activate
    def test_create(self) -> None:
        ssh_key_data: Dict[str, str] = {
            "sshKey": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQDf2pxWX/yhUBDyk2LPhvRtI0LnVO8PyR5Zt6AHrnhtLGqK+8YG9EMlWbCCWrASR+Q1hFQG example",
            "description": "Jim key"
        }
        self.client.ssh_keys.create(ssh_key_data)  # type: ignore

        assert len(responses.calls) == 1

    @responses.activate
    def test_update(self) -> None:
        ssh_key_id: int = 123
        ssh_key_data: Dict[str, str] = {
            "description": "Jim key"
        }

        try:
            self.client.ssh_keys.update(  # type: ignore
                ssh_key_id, ssh_key_data
            )
        except Exception as exc:
            assert False, f"'transip.TransIP.ssh_keys.update' raised an exception {exc}"



#!/usr/bin/python
# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages


def get_version() -> str:
    """
    Return the project version without having to import the module itself as it
    might not have been loaded yet.
    """
    with open("transip/__init__.py") as file:
        for line in file:
            if line.startswith("__version__"):
                return line.replace("\"", "").split()[-1]


def get_long_description() -> str:
    """
    Return the full description of the project. The full description consists
    of the combined README.md and CHANGELOG.md files.
    """
    with open("README.md", "r") as file:
        readme = file.read()
    with open("CHANGELOG.md", "r") as file:
        changelog = file.read()
    return readme + changelog


setup(
    name="python-transip",
    version=get_version(),
    description="Wrapper for the TransIP API",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Roald Nefs",
    author_email="info@roaldnefs.com",
    license="LGPLv3",
    url="https://github.com/roaldnefs/python-transip",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["cryptography>=3.3.1", "requests>=2.25.1"],
    python_requires=">=3.6",
    entry_points={},
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    extras_require={},
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020, 2021 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

import os
import base64

from typing import Optional, Type, List, Dict, Any

from transip.base import ApiService, ApiObject
from transip.mixins import (
    GetMixin, DeleteMixin, ListMixin, CreateMixin, UpdateMixin, ReplaceMixin,
    ObjectDeleteMixin, ObjectUpdateMixin,
    AttrsTuple
)
from transip.exceptions import TransIPIOError


class ApiTestService(ApiService):

    _path: str = "/api-test"

    def test(self):
        """
        A simple test to make sure everything is working.

        Returns:
            bool: True if everything is working, False otherwise.
        """
        response = self.client.get(f"{self.path}")
        if response.get('ping') == 'pong':
            return True
        return False


class ProductElement(ApiObject):

    _id_attr: str = "name"


class ProductElementService(ListMixin, ApiService):
    """Service to manage elements of a product."""

    _path: str = "/products/{parent_id}/elements"
    _obj_cls: Optional[Type[ApiObject]] = ProductElement

    _resp_list_attr: str = "productElements"


class Product(ApiObject):

    _id_attr: Optional[str] = "name"

    @property
    def elements(self) -> ProductElementService:
        """Return the service to manage the elements of the product."""
        return ProductElementService(
            self.service.client,
            parent=self  # type: ignore
        )


class ProductService(ListMixin, ApiService):
    """Service to manage products."""

    _path: str = "/products"
    _obj_cls: Optional[Type[ApiObject]] = Product

    _resp_list_attr: str = "products"

    def list(self) -> List[Type[ApiObject]]:
        """
        Retrieve a list of products.

        Overwrites the default list() method of the ListMixin as the products
        are stored in further down in the result dictionary.
        """
        objs: List[Type[ApiObject]] = []
        data = self.client.get(self.path)[self._resp_list_attr]
        # Loop over the individual product lists of all product categories,
        # e.g. vps, haip
        for obj_list in data.values():
            for obj in obj_list:
                objs.append(self._obj_cls(self, obj))  # type: ignore
        return objs


class AvailabilityZone(ApiObject):

    _id_attr: str = "name"


class AvailabilityZoneService(ListMixin, ApiService):

    _path: str = "/availability-zones"
    _obj_cls: Optional[Type[ApiObject]] = AvailabilityZone

    _resp_list_attr: str = "availabilityZones"


class SshKey(ObjectDeleteMixin, ObjectUpdateMixin, ApiObject):

    _id_attr: str = "id"


class SshKeyService(GetMixin, CreateMixin, UpdateMixin, DeleteMixin, ListMixin,
                    ApiService):

    _path: str = "/ssh-keys"
    _obj_cls: Optional[Type[ApiObject]] = SshKey

    _resp_list_attr: str = "sshKeys"
    _resp_get_attr: str = "sshKey"

    _create_attrs: Optional[AttrsTuple] = (
        ("sshKey",),  # required
        ("description",)  # optional
    )
    _update_attrs: Optional[AttrsTuple] = (
        ("description",),  # required
        tuple()  # optional
    )


class WhoisContact(ApiObject):

    _id_attr: Optional[str] = None


class WhoisContactService(ListMixin, ApiService):
    """Service to manage domain contacts of a domain."""

    _path: str = "/domains/{parent_id}/contacts"
    _obj_cls: Optional[Type[ApiObject]] = WhoisContact

    _resp_list_attr: str = "contacts"


class DnsEntry(ObjectUpdateMixin, ApiObject):

    _id_attr: Optional[str] = None

    def delete(self) -> None:
        """
        Delete a single DNS entry by calling the delete() method on its service
        and providing all the DNS entry attributes.

        This is different from the delete() method from the ObjectDeleteMixin
        as the deletion of a DNS entry requires all attributes of a single DNS
        entry and the DnsEntry does not have an ID.
        """
        self.service.delete(self.attrs)  # type: ignore

    def update(self) -> None:
        """
        Update the changes made to the DnsEntry.

        Overwrites the default update() method from the ObjectUpdateMixin
        because all attributes will need to be send when updating an DnsEntry
        and the DnsEntry does not have an ID.
        """
        updated_data = self._get_updated_data()
        if not updated_data:
            return

        self.service.update(updated_data)  # type: ignore


class DnsEntryService(CreateMixin, ListMixin, ReplaceMixin, ApiService):
    """Service to manage DNS entries of a domain."""

    _path: str = "/domains/{parent_id}/dns"
    _obj_cls: Optional[Type[ApiObject]] = DnsEntry

    # Additional data from creating a new single DNS entry using the
    # CreateMixin.
    _resp_list_attr: str = "dnsEntries"
    _req_create_attr: str = "dnsEntry"
    _create_attrs: AttrsTuple = (
        ("name", "expire", "type", "content"),  # required
        tuple()  # optional
    )

    # Additional data required to update a single DNS entry, the update()
    # method from the UpdateMixin can't be used as DNS entries don't have an
    # ID.
    _req_update_attr: str = "dnsEntry"
    _update_attrs: AttrsTuple = (
        ("name", "expire", "type", "content"),  # required
        tuple()  # optional
    )

    # Additional data required to delete a single DNS entry, the DeleteMixin
    # can't be used as DNS entries don't have an ID.
    _req_delete_attr: str = "dnsEntry"
    _delete_attrs: AttrsTuple = (
        ("name", "expire", "type", "content"),  # required
        tuple()  # optional
    )

    # Additional data from replacing all existing DnsEntry objects using the
    # ReplaceMixin.
    _req_replace_attr: str = "dnsEntries"
    _replace_attrs: AttrsTuple = (
        ("name", "expire", "type", "content"),  # required
        tuple()  # optional
    )

    def get_delete_attrs(self) -> AttrsTuple:
        """
        Return the required and optional attributes for deleting a DNS entry.

        Returns:
            tuple: a tuple containing a tuple of required and optional
                attributes.
        """
        return self._delete_attrs

    def get_update_attrs(self) -> AttrsTuple:
        """
        Return the required and optional attributes for updating a DNS entry.

        Returns:
            tuple: a tuple containing a tuple of required and optional
                attributes.
        """
        return self._update_attrs

    def _check_required_attrs(
        self,
        attrs: Dict[str, Any],
        expected_attrs: AttrsTuple,
    ) -> None:
        """
        Check if all the required attributes are included.

        Args:
            attrs: Dictionary containing the attributes to check.
            expected_attrs: The expected attributes in the form of a tuple
                containing a tuple with required and optional attributes.

        Raises:
            AttributeError: If any of the required attributes is missing.
        """
        required, optional = expected_attrs
        missing = [attr for attr in required if attr not in attrs]

        if missing:
            attrs_str = "', '".join(missing)
            raise AttributeError((
                f"'{self.__class__.__name__}' object has no "
                f"attribute{'s'[:len(missing)!=1]} '{attrs_str}'"
            ))

    def delete(self, data: Optional[Dict[str, Any]] = None) -> None:
        """
        Delete a DNS entry.

        This is different from the delete() method from the DeleteMixin as the
        deletion of a DNS entry requires all attributes of a single DNS entry
        and the DNS entries do not have an ID.
        """
        if data is None:
            data = {}

        # Check if all required attributes are supplied
        self._check_required_attrs(data, self.get_delete_attrs())

        # Requires the endpoint to be packed in dictionary with a specific key
        data = {self._req_delete_attr: data}

        if self.path:
            self.client.delete(f"{self.path}", json=data)

    def update(self, data: Optional[Dict[str, Any]] = None) -> None:
        """
        Update a single DnsEntry.

        This is different from the update() method from the UpdateMixin because
        a DnsEntry doesn't have a ID and the HTTP method needs to be PATCH
        instead of PUT.
        """
        if data is None:
            data = {}

        # Check if all required attributes are supplied.
        self._check_required_attrs(data, self.get_update_attrs())

        # Requires the endpoint to be packed in dictionary with a specific key
        data = {self._req_update_attr: data}

        if self.path:
            # Use the PATCH method to update a single DnsEntry.
            self.client.patch(f"{self.path}", json=data)


class Nameserver(ApiObject):

    _id_attr: Optional[str] = "hostname"


class NameserverService(ListMixin, ApiService):
    """Service to nameservers of a domain."""

    _path: str = "/domains/{parent_id}/nameservers"
    _obj_cls: Optional[Type[ApiObject]] = Nameserver

    _resp_list_attr: str = "nameservers"


class Domain(ApiObject):

    _id_attr: str = "name"

    @property
    def contacts(self) -> WhoisContactService:
        """Return the service to manage the WHOIS contacts of the domain."""
        return WhoisContactService(
            self.service.client,
            parent=self  # type: ignore
        )

    @property
    def dns(self) -> DnsEntryService:
        """Return the service to manage the DNS entries of the domain."""
        return DnsEntryService(
            self.service.client,
            parent=self  # type: ignore
        )

    @property
    def nameservers(self) -> NameserverService:
        """Return the service to manage the nameservers of the domain."""
        return NameserverService(
            self.service.client,
            parent=self  # type: ignore
        )


class DomainService(CreateMixin, GetMixin, DeleteMixin, ListMixin, ApiService):
    """Service to manage domains."""

    _path: str = "/domains"
    _obj_cls: Optional[Type[ApiObject]] = Domain

    _resp_list_attr: str = "domains"
    _resp_get_attr: str = "domain"

    _create_attrs: Optional[AttrsTuple] = (
        ("domainName",),  # required
        ("contacts", "nameservers", "dnsEntries")  # optional
    )


class InvoiceItem(ApiObject):
    pass


class InvoiceItemService(ListMixin, ApiService):
    """Service to items of an invoice."""

    _path: str = "/invoices/{parent_id}/invoice-items"
    _obj_cls: Optional[Type[ApiObject]] = InvoiceItem

    _resp_list_attr: str = "invoiceItems"


class Invoice(ApiObject):

    _id_attr: str = "invoiceNumber"

    @property
    def items(self) -> InvoiceItemService:
        """Return the service to manage the items of an invoice"""
        return InvoiceItemService(
            self.service.client,
            parent=self  # type: ignore
        )

    def pdf(self, file_path: str) -> Optional[str]:
        """
        Write a invoice to a PDF file.

        Args:
            file_path (str): Path to PDF file, if the path is a directory to
                PDF is saved using its invoice number.

        Returns:
            str: The absolute path to the saved PDF file.

        Raises:
            TransIPIOError: If the PDF data couldn't be written to file.
        """
        invoice_id = self.get_id()
        if not invoice_id:
            return None

        # Retrieve the base64 encoded PDF data
        encoded = self.service.client.get(f"/invoices/{invoice_id}/pdf")["pdf"]
        data = base64.b64decode(encoded.encode('ascii'))

        file_path = os.path.abspath(file_path)
        if os.path.isdir(file_path):
            file_path = os.path.join(file_path, f"{invoice_id}.pdf")
        if os.path.exists(file_path):
            raise TransIPIOError(f"File {file_path} already exists")

        try:
            with open(file_path, 'wb') as pdf_file:
                pdf_file.write(data)
        except OSError as exc:
            raise TransIPIOError(
                f"Unable to write PDF file {file_path}"
            ) from exc

        return file_path


class InvoiceService(GetMixin, ListMixin, ApiService):

    _path: str = "/invoices"
    _obj_cls: Optional[Type[ApiObject]] = Invoice

    _resp_list_attr: str = "invoices"
    _resp_get_attr: str = "invoice"


class Vps(ApiObject):

    _id_attr: str = "name"


class VpsService(GetMixin, DeleteMixin, ListMixin, ApiService):

    _path: str = "/vps"
    _obj_cls: Optional[Type[ApiObject]] = Vps

    _resp_list_attr: str = "vpss"
    _resp_get_attr: str = "vps"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Roald Nefs <info@roaldnefs.com>
#
# This file is part of python-transip.
#
# python-transip is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-transip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-transip.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional


class TransIPError(Exception):

    def __init__(self, message: str = "") -> None:
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        return self.message


class TransIPHTTPError(TransIPError):

    def __init__(
        self,
        message: str = "",
        response_code: Optional[int] = None
    ) -> None:

        super().__init__(message)
        self.response_code = response_code

    def __str__(self) -> str:
        if self.response_code:
            return f"{self.response_code}: {self.message}"
        else:
            return self.message


class TransIPParsingError(TransIPError):
    pass


class TransIPIOError(TransIPError):
    pass
